using System;
using System.Threading;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using static System.Diagnostics.Stopwatch;
using  GPUTools.Skinner.Scripts.Providers;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Common.Scripts.PL.Tools;
using Leap.Unity;


namespace Stopper {

	public class ClothCUA : MVRScript {
        List<GameObject> createdObjects = new List<GameObject>();

        List<ClothMeshControl> clothMeshControls = new List<ClothMeshControl>();

        List<MVRScriptExtensions.LayoutPanel> controlPanels = new List<MVRScriptExtensions.LayoutPanel>();

        Color buttonColor = new Color(0.839f, 0.839f, 0.839f);

        public class ClothMeshControl {
            public string name;
            public string storeId;
            public bool initialized;
            public MeshFilter meshFilter;
            public MeshRenderer meshRenderer;
            public Mesh mesh;

            public ClothSettings clothSettings;
            public CUADAZSkinV2 skin;

            public GameObject container;

            public float cuaScale;

            public MVRScriptExtensions.LayoutPanel controlPanel;
            public MVRScriptExtensions.TextBox status;

            public UIDynamicButton selectButton;

            public JSONStorableBool initOnStartup;

            
            public JSONStorableBool simEnabled;
            public JSONStorableBool accelerationEnabled;
            public JSONStorableBool collisionEnabled;

            public JSONStorableFloat iterations;
            
            public JSONStorableFloat uniformSimTextureStrength;

            public JSONStorableFloat stiffness;
            public JSONStorableFloat distanceScale;
            public JSONStorableFloat compressionResistance;
            public JSONStorableFloat skinJointStrength;

            public JSONStorableFloat gravityMultiplier;
            public JSONStorableFloat weight;
            public JSONStorableFloat drag;
            
            public JSONStorableFloat collisionRadius;
            public JSONStorableFloat collisionPower;
            public JSONStorableFloat friction;
            public JSONStorableFloat staticMultiplier;
            
            public void Enable() {
                if(this.initialized) {
                    this.meshRenderer.enabled = false;
                    this.clothSettings.enabled = true;
                    this.skin.enabled = true;
                }
            }

            public void Disable() {
                if(this.initialized) {
                    this.meshRenderer.enabled = true;
                    this.clothSettings.enabled = false;
                    this.skin.enabled = false;
                }
            }
        }

        public override void Init() {
            int storeSuffix = 1;
            foreach(MeshFilter meshFilter in containingAtom.reParentObject.GetComponentsInChildren<MeshFilter>()) {
                ClothMeshControl clothMeshControl = new ClothMeshControl();
                clothMeshControls.Add(clothMeshControl);

                clothMeshControl.name = $"{meshFilter.name}";
                clothMeshControl.storeId = $"{meshFilter.name}{storeSuffix++}";

                clothMeshControl.meshFilter = meshFilter;
                clothMeshControl.mesh = meshFilter.mesh;
                clothMeshControl.meshRenderer = meshFilter.GetComponent<MeshRenderer>();

                clothMeshControl.initOnStartup = new JSONStorableBool($"{clothMeshControl.storeId}InitOnStartup", false);
                clothMeshControl.initOnStartup.setCallbackFunction = (val) => {
                    
                };
                RegisterBool(clothMeshControl.initOnStartup);

                clothMeshControl.iterations = new JSONStorableFloat($"{clothMeshControl.storeId}Iterations", 3f, 1f, 7f);
                RegisterFloat(clothMeshControl.iterations);

                clothMeshControl.simEnabled = new JSONStorableBool($"{clothMeshControl.storeId}SimEnabled", true);
                RegisterBool(clothMeshControl.simEnabled);

                clothMeshControl.accelerationEnabled = new JSONStorableBool($"{clothMeshControl.storeId}AccelerationEnabled", true);
                RegisterBool(clothMeshControl.accelerationEnabled);

                clothMeshControl.collisionEnabled = new JSONStorableBool($"{clothMeshControl.storeId}CollisionEnabled", true);
                RegisterBool(clothMeshControl.collisionEnabled);

                clothMeshControl.uniformSimTextureStrength = new JSONStorableFloat($"{clothMeshControl.storeId}UniformStrength", 0.5f, 0f, 1f);
                RegisterFloat(clothMeshControl.uniformSimTextureStrength);


                clothMeshControl.stiffness = new JSONStorableFloat($"{clothMeshControl.storeId}Stiffness", 0.5f, 0f, 1f);
                RegisterFloat(clothMeshControl.stiffness);

                clothMeshControl.distanceScale = new JSONStorableFloat($"{clothMeshControl.storeId}DistanceScale", 1f, 0f, 1f);
                RegisterFloat(clothMeshControl.distanceScale);

                clothMeshControl.compressionResistance = new JSONStorableFloat($"{clothMeshControl.storeId}CompressionResistance", 0.5f, 0f, 1f);
                RegisterFloat(clothMeshControl.compressionResistance);

                clothMeshControl.skinJointStrength = new JSONStorableFloat($"{clothMeshControl.storeId}SkinJointStrength", 0.5f, 0f, 1f);
                RegisterFloat(clothMeshControl.skinJointStrength);

                clothMeshControl.gravityMultiplier = new JSONStorableFloat($"{clothMeshControl.storeId}GravityMultiplier", 0f, 0f, 1f);
                RegisterFloat(clothMeshControl.gravityMultiplier);

                clothMeshControl.weight = new JSONStorableFloat($"{clothMeshControl.storeId}Weight", 1f, 0f, 2f);
                RegisterFloat(clothMeshControl.weight);

                clothMeshControl.drag = new JSONStorableFloat($"{clothMeshControl.storeId}Drag", 0.2f, 0f, 1f);
                RegisterFloat(clothMeshControl.drag);

                clothMeshControl.collisionRadius = new JSONStorableFloat($"{clothMeshControl.storeId}CollisionRadius", 0.01f, 0f, 0.2f);
                RegisterFloat(clothMeshControl.collisionRadius);

                clothMeshControl.collisionPower = new JSONStorableFloat($"{clothMeshControl.storeId}CollisionPower", 1f, 0f, 1f);
                RegisterFloat(clothMeshControl.collisionPower);

                clothMeshControl.friction = new JSONStorableFloat($"{clothMeshControl.storeId}Friction",0.5f, 0f, 1f);
                RegisterFloat(clothMeshControl.friction);

                clothMeshControl.staticMultiplier = new JSONStorableFloat($"{clothMeshControl.storeId}StaticMultiplier", 2f, 0f, 1f);
                RegisterFloat(clothMeshControl.staticMultiplier);
            }

        }

		private void Start(){
            InitMainUI();
		}

        private void InitMainUI() {
            var mainUI = this.rightUIContent.transform.parent;
            var mainPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("MainPanel", mainUI);

            var headerPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("HeaderPanel", mainPanel);

            var cuaControlPanel = this.AddUIElement<MVRScriptExtensions.VerticalSplitLayoutPanel>("CUAControlPanel", mainPanel);
            
            foreach(var clothMeshControl in clothMeshControls) {
                clothMeshControl.selectButton = CreateButton($"{clothMeshControl.name}");
                clothMeshControl.selectButton.button.onClick.AddListener(() => {
                    foreach(var cmc in clothMeshControls) {
                        cmc.controlPanel.gameObject.SetActive(false);
                        cmc.selectButton.buttonColor = buttonColor;
                    }
                    clothMeshControl.controlPanel.gameObject.SetActive(true);
                    clothMeshControl.selectButton.buttonColor = Color.green;
                });

                clothMeshControl.controlPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>($"{clothMeshControl.storeId}ControlPanel", cuaControlPanel.rightPanel);
                controlPanels.Add(clothMeshControl.controlPanel);

                var header = this.AddTextBox($"{clothMeshControl.name}", clothMeshControl.controlPanel);
                header.UIText.fontStyle = FontStyle.Bold;
                header.UIText.alignment = TextAnchor.UpperLeft;

                clothMeshControl.status = this.AddTextBox($"", clothMeshControl.controlPanel);
                clothMeshControl.status.UIText.alignment = TextAnchor.MiddleLeft;
                clothMeshControl.status.gameObject.SetActive(false);
              
                var initClothButton = CreateButton("Create Cloth");
                initClothButton.button.onClick.AddListener(() => {
                    StartCoroutine(CreateCUACloth(clothMeshControl));

                    initClothButton.gameObject.SetActive(false);
                    clothMeshControl.status.gameObject.SetActive(true);
                });
                this.AddToPanel(initClothButton, clothMeshControl.controlPanel);

                clothMeshControl.controlPanel.gameObject.SetActive(false);
            }
           // this.AddToPanel(simProxyChooserPopup, selectionPanel);
        }

        

        private IEnumerator CreateCUACloth(ClothMeshControl control) {
            control.container = new GameObject($"ClothCUA{control.storeId}");
            control.container.layer = 0;

            Transform parent = containingAtom.reParentObject.Find("object/rescaleObject");
            if(parent.childCount > 0) {
                parent = parent.GetChild(0);
            }
            control.container.transform.SetParent(parent, false);
            // control.container.transform.localScale = parent.localScale;
            createdObjects.Add(control.container);
            control.container.SetActive(false);

            DAZMeshData meshData = ScriptableObject.CreateInstance<DAZMeshData>();
            meshData.baseVertices = (Vector3[])control.mesh.vertices.Clone();
            meshData.UVVertices = meshData.baseVertices;
            meshData.OrigUV = (Vector2[])control.mesh.uv.Clone();

            List<DAZVertexMap> baseVerticesToUV = new List<DAZVertexMap>();
            for(int i = 0; i < meshData.baseVertices.Length; i++) {
                DAZVertexMap vm = new DAZVertexMap();
                vm.fromvert = i;
                vm.tovert = i;
                baseVerticesToUV.Add(vm);
            }
            meshData.baseVerticesToUVVertices = baseVerticesToUV.ToArray();

            List<MeshPoly> meshPolies = new List<MeshPoly>();
            var triangles = control.mesh.triangles;
            for(int i = 0; i < triangles.Length/3; i++) {
                MeshPoly meshPoly = new MeshPoly();
                meshPoly.materialNum = 0;
                meshPoly.vertices = new int[]{triangles[3*i+2],triangles[3*i+1],triangles[3*i]};
                meshPolies.Add(meshPoly);
            }

            meshData.basePolyList = meshPolies.ToArray();
            meshData.UVPolyList = meshData.basePolyList;

            
            Atom person = SuperController.singleton.GetAtoms().First(a => a.type == "Person");
            DAZSkinV2 personSkin = person.GetComponentInChildren<DAZSkinV2>();

            CUADAZMesh dazMesh = control.container.AddComponent<CUADAZMesh>();
            dazMesh.meshData = meshData;
            dazMesh.materials = new Material[control.meshRenderer.materials.Length];
            for(int i = 0; i < dazMesh.materials.Length; i++) {
                dazMesh.materials[i] = new Material(control.meshRenderer.materials[i]);
            }
          
            dazMesh.UpdateInternals();
            dazMesh.Init();

            control.meshRenderer.enabled = false;

            dazMesh.drawUVMappedMesh = false;
            
            CUADAZSkinV2 skin = control.container.AddComponent<CUADAZSkinV2>();
            control.skin = skin;
            skin.dazMesh = dazMesh;
            skin.skin = false;
            skin.draw = false;
            skin.drawInPostProcess = true;
            skin.useGeneralWeights = true;
            skin.useSmoothing = false;
            skin.allowPostSkinMorph = false;
            skin.GPUSkinner = Instantiate(personSkin.GPUSkinner);
            skin.GPUMeshCompute = Instantiate(personSkin.GPUMeshCompute);
            skin.useEarlyFinish = true;
            skin.skinMethod = DAZSkinV2.SkinMethod.GPU;
            skin.CopyMaterials();
            skin.InitSimpleBones();
            foreach(var mat in skin.GPUmaterials) {
                mat.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
            }

            
            var pos = control.container.transform.position;
            var rot = control.container.transform.rotation;
            control.container.transform.position = Vector3.zero;
            control.container.transform.rotation = Quaternion.identity;
            control.container.SetActive(true);

            control.container.transform.position = pos;
            control.container.transform.rotation = rot;

            // yield return new WaitForFixedUpdate();

            var clothSettings = control.container.AddComponent<ClothSettings>();
            control.clothSettings = clothSettings;
            clothSettings.enabled = false;
            clothSettings.EditorType = ClothEditorType.Texture;
            clothSettings.EditorStrengthType = ClothEditorType.Texture;
            
            clothSettings.MeshProvider.Type = ScalpMeshType.PreCalc;
            clothSettings.MeshProvider.PreCalcProvider = skin;

            control.cuaScale = control.container.transform.parent.localScale.x;
            clothSettings.WorldScale = control.container.transform.parent.localScale.x;

            clothSettings.ProcessGeometryMainThread();
            var geometryThread = new Thread(new ThreadStart(clothSettings.ProcessGeometryThreaded));
            geometryThread.Start();
            while(geometryThread.IsAlive) {
                control.status.text = clothSettings.GeometryData.status;
                yield return null;
            }

            control.status.gameObject.SetActive(false);
            clothSettings.CollisionEnabled = true;
            clothSettings.enabled = true;
            
            control.initialized = true;

            SyncControlUI(control);

            yield return null;
            clothSettings.Reset();
            yield break;
        }

        private void SyncControlUI(ClothMeshControl control) {
            var clothSettings = control.clothSettings;

            var syncMaterialsButton = CreateButton("Sync Materials");
            syncMaterialsButton.button.onClick.AddListener(() => {
                for(int i = 0; i < control.skin.GPUmaterials.Length; i++) {
                    control.skin.GPUmaterials[i] = new Material(control.meshRenderer.materials[i]);
                    control.skin.GPUmaterials[i].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
                    control.skin.FlushBuffers();
                }
            });
            this.AddToPanel(syncMaterialsButton, control.controlPanel);


            var resetButton = CreateButton("Reset");
            resetButton.button.onClick.AddListener(() => clothSettings.Reset());
            this.AddToPanel(resetButton, control.controlPanel);

            var simEnabledToggle = CreateToggle(control.simEnabled);
            simEnabledToggle.label = "Sim Enabled";
            this.AddToPanel(simEnabledToggle, control.controlPanel);
            control.simEnabled.setCallbackFunction = val => {
                if(val) {
                    control.Enable();
                } else {
                    control.Disable();
                }
            };

            var accelerationToggle = CreateToggle(control.accelerationEnabled);
            accelerationToggle.label = "Acceleration Enabled";
            this.AddToPanel(accelerationToggle, control.controlPanel);
            control.accelerationEnabled.setCallbackFunction = val => clothSettings.IntegrateEnabled = val;

            var collisionToggle = CreateToggle(control.collisionEnabled);
            collisionToggle.label = "Collision Enabled";
            this.AddToPanel(collisionToggle, control.controlPanel);
            control.collisionEnabled.setCallbackFunction = val => clothSettings.CollisionEnabled = val;

            var iterationsSlider = AddSliderForMeshControl(control.iterations, "Solver Iterations", control);
            iterationsSlider.slider.wholeNumbers = true;

            var stiffnessSlider = AddSliderForMeshControl(control.stiffness, "Stiffness", control);
            var distanceScale = AddSliderForMeshControl(control.distanceScale, "Distance Scale", control);
            var compressionResistance = AddSliderForMeshControl(control.compressionResistance, "Compression Resistance", control);
            var uniformSimTextureStrength = AddSliderForMeshControl(control.uniformSimTextureStrength, "Sim Texture Strength", control);
            var skinJointStrength = AddSliderForMeshControl(control.skinJointStrength, "Skin Joint Strength", control);
            var gravityMultiplier = AddSliderForMeshControl(control.gravityMultiplier, "Gravity Multiplier", control);
            var weight = AddSliderForMeshControl(control.weight, "Weight", control);
            var drag = AddSliderForMeshControl(control.drag, "Drag", control);
            var collisionRadius = AddSliderForMeshControl(control.collisionRadius, "Collision Radius", control);
            var collisionPower = AddSliderForMeshControl(control.collisionPower, "Collision Power", control);
            var friction = AddSliderForMeshControl(control.friction, "Friction", control);
            var staticMultiplier = AddSliderForMeshControl(control.staticMultiplier, "Static Multiplier", control);

            SetCallbackFunction(control.uniformSimTextureStrength, val => {
                int num = 4;
                int num2 = 4;
                Color color = new Color(val, 0f, 0f);
                Texture2D texture2D = new Texture2D(num, num2);
                for (int i = 0; i < num; i++)
                {
                    for (int j = 0; j < num2; j++)
                    {
                        texture2D.SetPixel(i, j, color);
                    }
                }
                texture2D.Apply();

                clothSettings.EditorTexture = texture2D;

                if (clothSettings.GeometryData != null)
				{
					clothSettings.GeometryData.ResetParticlesBlend();
				}
				if (clothSettings.builder != null)
				{
					if (clothSettings.builder.physicsBlend != null)
					{
						clothSettings.builder.physicsBlend.UpdateSettings();
					}
					clothSettings.Reset();
				}                   
            });       

            SetCallbackFunction(control.stiffness, val => clothSettings.Stiffness = val);           

            SetCallbackFunction(control.distanceScale, val => clothSettings.DistanceScale = val);

            SetCallbackFunction(control.compressionResistance, val => clothSettings.CompressionResistance = val);

            SetCallbackFunction(control.skinJointStrength, val => clothSettings.JointStrength = val);

            SetCallbackFunction(control.gravityMultiplier, val => clothSettings.GravityMultiplier = val);

            SetCallbackFunction(control.weight, val => clothSettings.Weight = val);

            SetCallbackFunction(control.drag, val => clothSettings.Drag = val);

            SetCallbackFunction(control.collisionRadius, val => {clothSettings.ParticleRadius = val; clothSettings.Reset();});

            SetCallbackFunction(control.collisionPower, val => clothSettings.CollisionPower = val);

            SetCallbackFunction(control.friction, val => clothSettings.Friction = val);

            SetCallbackFunction(control.staticMultiplier, val => clothSettings.StaticMultiplier = val);
        }

        private UIDynamicSlider AddSliderForMeshControl(JSONStorableFloat storableFloat, string label, ClothMeshControl control) {
            var slider = CreateSlider(storableFloat);
            slider.label = label;
            this.AddToPanel(slider, control.controlPanel);
            return slider;
        }

        private void SetCallbackFunction(JSONStorableFloat storableFloat, JSONStorableFloat.SetFloatCallback callback) {
            storableFloat.setCallbackFunction = callback;
            storableFloat.setCallbackFunction(storableFloat.val);
        }
		
        public void OnDisable(){	
            foreach(var control in clothMeshControls) {
                control.Disable();
            }
        }

        public void OnEnable(){	
            foreach(var control in clothMeshControls) {
                control.Enable();
            }
        }

        public void OnDestroy(){
            foreach(var go in createdObjects) {
                Destroy(go);
            }
        }
        
        public class CUADAZMesh : DAZMesh {
            
            public void UpdateInternals() {
                this._numBaseVertices = this.meshData.baseVertices.Length;
                this._numUVVertices = this.meshData.UVVertices.Length;
                this._numBasePolygons = this.meshData.basePolyList.Length;
                this._numMaterials = this.materials.Length;

                this.materialsEnabled = new bool[this.materials.Length];
                this.materialsShadowCastEnabled = new bool[this.materials.Length];
                this._materialNames = new string[this.materials.Length];
                for(int i = 0; i < this.materials.Length; i++) {
                    this.materialsEnabled[i] = true;
                    this.materialsShadowCastEnabled[i] = true;
                    this._materialNames[i] = this.materials[i].name;
                }
            }
        }

        public class CUADAZSkinV2 : DAZSkinV2 {

            public void InitSimpleBones() {
                GameObject rootBoneGo = new GameObject("root");
                rootBoneGo.transform.SetParent(this.transform,true);
                rootBoneGo.transform.localPosition = Vector3.zero;
                rootBoneGo.transform.localRotation = Quaternion.identity;
                
                DAZBone rootBone = rootBoneGo.AddComponent<DAZBone>();
                rootBone.enabled = true;
                rootBone.isRoot = true;

                DAZBones root = this.gameObject.AddComponent<DAZBones>();
                root.dazBones = new DAZBone[]{rootBone};
                // root.useScale = true;
                
                DAZSkinV2Node node = new DAZSkinV2Node();
                node.name = rootBone.name;
                node.url = node.name;
                node.id = node.name;

                // We'll use general weights
                node.weights = new DAZSkinV2VertexWeights[0];
                node.fullyWeightedVertices = new int[0];
                node.generalWeights = new DAZSkinV2GeneralVertexWeights[this.dazMesh.numBaseVertices];
                for(int i = 0; i < node.generalWeights.Length; i++) {
                    DAZSkinV2GeneralVertexWeights generalWeight = new DAZSkinV2GeneralVertexWeights();
                    generalWeight.vertex = i;
                    generalWeight.weight = 1;
                    node.generalWeights[i] = generalWeight;
                }
                
                DAZSkinV2BulgeFactors bulgeFactors = new DAZSkinV2BulgeFactors();
                bulgeFactors.name = "root";
                node.bulgeFactors = bulgeFactors;

                nodes = new DAZSkinV2Node[]{node};
                rootBone.dazBones = root;
                this.dazBones = new DAZBone[] {rootBone};
                this.root = root;
                this._numBones = 1;
            }

            public override Mesh MeshForImport {
                get {
                    return Mesh;
                }
            }

            public override GpuBuffer<Vector3> PreCalculatedVerticesBuffer {
                get {
                    var buffer = base.PreCalculatedVerticesBuffer;
                    if(buffer.Data == null) {
                        buffer = new GpuBuffer<Vector3>(this.rawVertsBuffer);
                        this.PreCalculatedVerticesBuffer = buffer;
                        buffer.Data = new Vector3[rawVertsBuffer.count];
                    }
                    return buffer;
                }
                protected set {
                    base.PreCalculatedVerticesBuffer = value;
                }
            }


            // Overwriting this to use the rawvertsbuffer
            public override void Dispatch() {
                if (this.needsDispatch && this.rawVertsBuffer != null) {
                    this.SkinMeshGPU();
                    this.needsDispatch = false;
                    this.GPUSkinner.SetBuffer(this._copyKernel, "inVerts", this.rawVertsBuffer);
                    this.GPUSkinner.SetBuffer(this._copyKernel, "outVerts", this.preCalcVertsBuffer);
                    this.GPUSkinner.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                    if (this.provideToWorldMatrices) {
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalPositions", this._originalVerticesBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalNormals", this._originalNormalsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalTangents", this._originalTangentsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentPositions", this.preCalcVertsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentNormals", this._normalsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentTangents", this._tangentsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "vertexChangeMatrices", this._matricesBuffer);
                        this.GPUSkinner.Dispatch(this._calcChangeMatricesKernel, this.numVertThreadGroups, 1, 1);
                    }
                }
            }


            public override void PostProcessDispatch(ComputeBuffer finalVerts) {
                if (this.drawInPostProcess && this.preCalcVertsBuffer != null) {
                    this.GPUSkinner.SetBuffer(this._copyKernel, "inVerts", finalVerts);
                    this.GPUSkinner.SetBuffer(this._copyKernel, "outVerts", this.rawVertsBuffer);
                    this.GPUSkinner.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                   
                    this.DrawMeshGPU();
                }
            }
        }
    }
}